from typing import *

COLOR_NAME_TO_HEX_MAP: Dict[str, int] = {
    "Blue": 0xAECBFA,
    "Brown": 0xE6C9A8,
    "DarkBlue": 0xAECBFA,
    "Gray": 0xE8EAED,
    "Green": 0xCCFF90,
    "Orange": 0xFBBC04,
    "Pink": 0xFDCFE8,
    "Purple": 0xD7AEFB,
    "Red": 0xF28B82,
    "Teal": 0xA7FFEB,
    "White": 0xFFFFFF,
    "Yellow": 0xFFF475,
}

# colors
C_PRIMARY = 0x268CCE



from ideaseed.utils import dye

DUMB_UTF8_ART = """


                ██████████  ██  ██  
                ██      ██  ██  ██  
                ██████████  ██  ██ 
                ██      ██  ██  ██ 
                ██████████████████  <- that's my personal logo,
                ██      ██      ██     i didn't make one for 
                ██████████      ██     ideaseed, yet™
                ██      ██      ██ 
                ██████████      ██ 
      
            ideaseed v{version}
            by Ewen Le Bihan
            more at https://ewen.works
      
                ~ thx to these ppl ~

https://github.com/kiwiz
    This madman reverse-engineered Google Keep's
    internal REST API so that anyone could 
    use it with ease.
                    
https://github.com/PyGithub
    This one of the cleanest libs I've ever used.
    Like `requests`-level cleanliness.
                    
https://github.com/docopt
    Designing CLIs with this is a fucking breeze.
    I can almost copy-paste documentation into
    a docstring and call it a day, it's crazy.
"""



import os
import pprint
from os.path import dirname
import github
from github import Issue
from github.GithubException import BadCredentialsException, TwoFactorException
from ideaseed.utils import ask, dye, get_token_cache_filepath
from ideaseed.constants import C_PRIMARY
from typing import *
from github import Github
import json
from github.GithubException import TwoFactorException
import inquirer as q
from ideaseed.utils import ask


def login_with_cache() -> Optional[Github]:
    """
    Tries to login using the cached credentials.
    `None` is returned if the cache does not exist or is invalid.
    """
    if os.path.exists(get_token_cache_filepath('github')):
        with open(get_token_cache_filepath('github'), encoding="utf8") as file:
            creds = json.load(file)
            creds = (
                [creds["username"], creds["password"]]
                if creds["pat"] is None
                else [creds["pat"]]
            )
            try:
                gh = Github(*creds)
                gh.get_user().name
                return gh
            except BadCredentialsException:
                return None
            except TwoFactorException:
                return None
    return None


def clear_auth_cache():
    if os.path.exists(get_token_cache_filepath('github')):
        os.remove(get_token_cache_filepath('github'))


def write_auth_cache(auth):
    os.makedirs(dirname(get_token_cache_filepath('github')), exist_ok=True)
    with open(get_token_cache_filepath('github'), "w", encoding="utf8") as file:
        json.dump(auth, file)


def login(args: Dict[str, Any], method: Optional[str] = None) -> Github:
    """
    Returns a `Github` instance to interact with.
    Prompts the user to login, either via username/password
    or using a Personal Access Token
    """
    gh = login_with_cache()
    if gh is not None:
        return gh
    else:
        del gh

    questions = [
        q.List(
            name="method",
            message="Log in using",
            choices=["Personal Access Token", "Username and password"],
            ignore=lambda _: method is not None,
        ),
        q.Text(
            name="pat",
            message="Personal Access Token",
            ignore=lambda ans: (method or ans["method"]) != "Personal Access Token",
        ),
        q.Text(
            name="username",
            message="Username",
            ignore=lambda ans: (method or ans["method"]) != "Username and password",
        ),
        q.Password(
            name="password",
            message="Password",
            ignore=lambda ans: (method or ans["method"]) != "Username and password",
        ),
    ]
    answers = q.prompt(questions)

    if not args["--no-auth-cache"]:
        write_auth_cache(answers)

    if answers["pat"] is not None:
        try:
            gh = Github(answers["pat"])
            github_username(gh)
            return gh
        except BadCredentialsException:
            print("Bad token")
            return login(args, method)
    else:
        try:
            gh = Github(answers["username"], answers["password"])
            github_username(gh)
            return gh
        except TwoFactorException:
            print(
                "Your account uses two-factor authentification. Please use a personal access token instead."
            )
            return login(args, method="Personal Access Token")
        except BadCredentialsException:
            print("Bad credentials")
            return login(args, method)


def github_username(gh: Github) -> str:
    return gh.get_user().login


def resolve_self_repository_shorthand(gh: Github, repo: str) -> str:
    if "/" not in repo:
        return github_username(gh) + "/" + repo
    return repo


def push_to_repo(args: Dict[str, Any]) -> None:
    gh = login(args)
    repo_name = resolve_self_repository_shorthand(gh, args["REPO"])
    idea = args["IDEA"]
    project_name = args["PROJECT"]
    column_name = args["COLUMN"]
    print(
        f"Saving card in {dye(repo_name, C_PRIMARY)}"
        f" › {dye(project_name, C_PRIMARY)}"
        f" › {dye(column_name, C_PRIMARY)}..."
    )
    repo = gh.get_repo(repo_name)

    project = None
    for p in repo.get_projects():
        if p.name.lower() == project_name.lower():
            project = p
            break

    # Project not found
    if project is None and args["--create-missing"]:
        if ask(q.Confirm("ans", message=f"Create missing project {project_name!r}?")):
            description: str = ask(
                q.Editor("ans", message="Enter the project's description...")
            )
            project = repo.create_project(name=project_name, body=description)
        else:
            return
    # Not found and not create
    elif project is None:
        print(dye(f"Error: project {project_name!r} does not exist!", fg=0xF00))
        return

    # Column
    column = None
    for c in project.get_columns():
        if c.name.lower() == column_name.lower():
            column = c
            break

    # Column not found
    if column is None and args["--create-missing"]:
        if ask(q.Confirm("ans", message=f"Create missing column {column_name!r}?")):
            column = project.create_column(column_name)
        else:
            return
    # Not found and not create
    elif column is None:
        print(dye(f"Error: column {column_name!r} does not exist!", fg=0xF00))
        return

    if args["--issue"]:
        issue = repo.create_issue(
            title=args["--title"] or idea,
            body=idea if args["--title"] else "",
            assignees=[github_username(gh)],
            labels=args["--tag"],
        )
        card = column.create_card(content_id=issue.id, content_type="Issue")
    else:
        card = column.create_card(note=idea)
        pass

def push_to_user(args: Dict[str, Any]) -> None:
    gh = login(args)
    idea = args["IDEA"]
    project_name: str = args["--user-project"]
    column_name: str = args["PROJECT"]
    print(
        f"Saving card in {dye(github_username(gh), C_PRIMARY)} › {dye(project_name, C_PRIMARY)} › {dye(column_name, C_PRIMARY)}..."
    )
    project = None
    user = gh.get_user()
    for p in user.get_projects():
        if p.name.lower() == project_name.lower():
            project = p
            break

    # Project not found
    if project_name is None and args["--create-missing"]:
        if ask(q.Confirm("ans", message=f"Create missing project {project_name!r}?")):
            description = ask(
                q.Editor("ans", message="Enter the project's description...")
            )
            project = user.create_project(name=project_name, body=description)
        else:
            return
    # Not found and not create
    elif project_name is None:
        print(dye(f"Error: project {project_name!r} does not exist!", fg=0xF00))
        return

    # Column
    column = None
    for c in project.get_columns():
        if c.name.lower() == column_name.lower():
            column = c
            break

    # Column not found
    if column_name is None and args["--create-missing"]:
        if ask(q.Confirm("ans", message=f"Create missing column {column_name!r}?")):
            column = project.create_column(column_name)
        else:
            return
    # Not found and not create
    elif column_name is None:
        print(dye(f"Error: column {column_name!r} does not exist!", fg=0xF00))
        return

    column.create_card(note=idea)



"""Note down your ideas and get them to the right place, without switching away from your terminal

Usage: 
    ideaseed (--help | --about | --version)
    ideaseed [options] [-t TAG...] ARGUMENTS...

Examples:
    # Save a card "test" in schoolsyst/webapp > project "UX" > column "To-Do"
    $ ideaseed schoolsyst/webapp UX "test"
    # Save a card "lorem" in your-username/ipsum > project "ipsum" > column "To-Do"
    $ ideaseed ipsum "lorem"
    # Save a card "a CLI to note down ideas named ideaseed" in your user profile > project "incubator" > column "willmake"
    $ ideaseed --user-keyword=project --user-project=incubator project "a CLI to note down ideas named ideaseed"

Arguments:
    REPO        Select a repository
                    If not given                        Uses Google Keep instead of GitHub (or uses your user profile's projects if --project is used)
                    If --user-keyword's value is given  Creates a card on your user's project (select which project with --user-project)
                    If given in the form OWNER/REPO     Uses the repository OWNER/REPO
                    If given in the form REPO           Uses the repository "your username/REPO"
    PROJECT     Select a project by name to put your card to [default: REPO's value]
                    If creating a card on your user's project, this becomes the COLUMN
    COLUMN      Select a project's column by name [default: To do]
                    If creating a card on your user's project, this is ignored

Options:
    -c --color COLOR        Chooses which color to use for Google Keep cards. See Color names for allowed values.
    -t --tag TAG            Adds tags to the Google Keep card. Use it multiple times to set multiple tags.
                            When used together with --issue, --tag means labels.
    -i --issue              Creates an issue for the card and link them together. IDEA becomes the issue's title, except if --title is specified,
                            in which case IDEA becomes the issue's description and --title's value the issue title.
    (WIP) -I --interactive  Prompts you for the above options when they are not provided.
    -T --title TEXT         Sets the Google Keep card's title. When used with --issue, sets the issue's title.
    -L --logout             Clears the authentification cache
    -m --create-missing     Create non-existant tags, labels, projects or columns specified, upon confirmation.
       --about              Details about ideaseed like currently-installed version
       --version            Like --about, without dumb and useless stuff

Settings options: It's comfier to set these in your alias, e.g. alias idea="ideaseed --user-project=incubator --user-keyword=project --no-auth-cache --create-missing"
       --user-project NAME  Name of the project to use as your user project
       --user-keyword NAME  When REPO is NAME, creates a GitHub card on your user profile instead of putting it on REPO
       --no-auth-cache      Don't save credentials in a temporary file at {cache_filepath}

Color names: Try with the first letter only too
    blue, brown, darkblue, gray, green, orange, pink, purple, red, teal, white, yellow
"""

from ideaseed.gkeep import push_to_gkeep
from ideaseed.github import clear_auth_cache, push_to_repo, push_to_user
from typing import *
from docopt import docopt
from pprint import pprint
from ideaseed.utils import dye, get_token_cache_filepath
from ideaseed.constants import COLOR_NAME_TO_HEX_MAP
from ideaseed.dumb_utf8_art import DUMB_UTF8_ART


def run(argv=None):
    args = resolve_arguments(
        docopt(__doc__.format(cache_filepath=get_token_cache_filepath("*")), argv)
    )
    validate_argument_presence(args)
    args = resolve_arguments_defaults(args)

    if args["--about"]:
        print(DUMB_UTF8_ART.format(version="0.1.0"))
        return

    if args["--version"]:
        print("0.1.0")
        return

    if args["--color"]:
        args["--color"] = expand_color_name(args["--color"])

    if args["--logout"]:
        clear_auth_cache()

    if args["REPO"]:
        if args["REPO"] == args["--user-keyword"]:
            push_to_user(args)
        else:
            push_to_repo(args)
    else:
        push_to_gkeep(args)


def resolve_arguments(args: Dict[str, Any]) -> Dict[str, Any]:
    """
    Apparently required positional arguments must come first...
    So I'm resolving them by hand here.
    """
    positional_args = args["ARGUMENTS"]
    idea = repo = project = column = None
    if len(positional_args) == 1:
        idea = positional_args[0]
    if len(positional_args) == 2:
        repo, idea = positional_args
    if len(positional_args) == 3:
        repo, project, idea = positional_args
    if len(positional_args) >= 4:
        repo, project, column, idea = positional_args
    return {"REPO": repo, "PROJECT": project, "COLUMN": column, "IDEA": idea, **args}


def resolve_arguments_defaults(args: Dict[str, Any]) -> Dict[str, Any]:
    return {
        **args,
        "PROJECT": args["PROJECT"] or args["REPO"],
        "COLUMN": args["COLUMN"] or "To do",
    }


class ValidationError(Exception):
    pass


def validate_argument_presence(args: Dict[str, str]) -> None:
    """
    Raises a `ValidationError` if one of the arguments is not allowed based
    on the other arguments.
    """

    GOOGLE_KEEP_ONLY = ("--color", "--tag")
    GITHUB_ONLY = ("--issue",)
    using_github = len(args["ARGUMENTS"]) > 1

    if using_github and any([v for k, v in args.items() if k in GOOGLE_KEEP_ONLY]):
        raise ValidationError(
            "The following options are not allowed when using GitHub: "
            + ", ".join(GOOGLE_KEEP_ONLY)
            + f"\n{args!r}"
        )
    if not using_github and any([v for k, v in args.items() if k in GITHUB_ONLY]):
        raise ValidationError(
            "The following options are not allowed when using Google Keep: "
            + ", ".join(GITHUB_ONLY)
        )


def expand_color_name(color: str) -> str:
    # All possible color names
    color_names = [k for k in COLOR_NAME_TO_HEX_MAP.keys()]
    # Initialize the array of matches
    matching_color_names: List[str] = []
    # Filter `color_names` to only get the color names that start with `color`
    for color_name in color_names:
        if color_name.lower().startswith(color.lower()):
            matching_color_names += [color_name]
    # If we have exactly _one_ element of `color_names` that matches, we return the only
    # element: it's a match!
    if len(matching_color_names) == 1:
        return matching_color_names[0]
    # If we have multiple choices, the given color was ambiguous.
    if len(matching_color_names) > 1:
        matching_color_names_display: List[str] = []
        for matching_color_name in matching_color_names:
            matching_color_names_display += [
                dye(
                    matching_color_name,
                    fg=0x000,
                    bg=COLOR_NAME_TO_HEX_MAP[matching_color_name],
                )
            ]
        raise ValueError(
            f"Ambiguous color shorthand {color!r}, it could mean one of: "
            + ", ".join(matching_color_names_display)
        )
    # If we had an empty array, then the given color name was incorrect.
    raise ValueError(f"Unknown color {color}")



from ideaseed.constants import COLOR_NAME_TO_HEX_MAP, C_PRIMARY
import json
from os import path
from ideaseed.utils import ask, dye, get_token_cache_filepath
from typing import *
from inquirer import Text, Password, Confirm
from gkeepapi import Keep
from gkeepapi.exception import LoginException
from gkeepapi.node import ColorValue
import sys


def write_to_cache(keep: Keep, email: str) -> None:
    with open(get_token_cache_filepath("gkeep"), "w", encoding="utf8") as file:
        json.dump({"master_token": keep.getMasterToken(), "email": email}, file)


def login_from_cache() -> Optional[Keep]:
    if not path.exists(get_token_cache_filepath("gkeep")):
        return None

    with open(get_token_cache_filepath("gkeep"), encoding="utf8") as file:
        creds = json.load(file)

    keep = Keep()
    keep.resume(email=creds["email"], master_token=creds["master_token"])
    return keep


def login(args: Dict[str, Any]) -> Keep:
    # Try to log in from cache
    keep = login_from_cache()
    if keep is not None:
        return keep
    else:
        del keep

    # Ask for creds
    username, password = ask(
        Text("u", message="E-mail"), Password("p", message="Password")
    )

    # Log in
    keep = Keep()
    try:
        keep.login(username, password)
        write_to_cache(keep, username)
    except LoginException as error:
        # Handle errors...
        (topic, message) = error.args
        if topic == "BadAuthentification":
            print(dye("Bad credentials", fg=0xF00))
            return login(args)
        elif topic == "NeedsBrowser":
            print(
                dye(
                    """You have two-step authentification set up, please add an App Password.
Go to https://myaccount.google.com/apppasswords,
Click on 'Generate', Choose a name and a device, then copy the code
and use it as your password.""",
                    fg=0xF00,
                )
            )
            sys.exit()
    return keep


def push_to_gkeep(args: Dict[str, Any]) -> None:
    # Log in
    print("Logging in...", end="")
    keep = login(args)
    print(" Done.")

    # Announce what we'll do
    color = args["--color"] or 'white'
    print(
        f"Creating a "
        + dye(' ' + color.lower() + ' ', fg=0x000, bg=COLOR_NAME_TO_HEX_MAP[color])
        + dye(f" Google Keep card with tags {', '.join(args['--tag'])}", fg=C_PRIMARY)
    )

    # Create the note
    note = keep.createNote(title=args["--title"], text=args["IDEA"])
    note.color = getattr(ColorValue, args["--color"])
    note.pinned = True

    # Find/create all the labels
    all_tags = keep.labels()
    for tag in args["--tag"]:
        label = keep.findLabel(tag)
        if label is None and args["--create-missing"]:
            if ask(Confirm("ans", f"Create missing tag {tag!r}?")):
                label = keep.createLabel(tag)
        elif label is None:
            print(dye(f"Error: Tag {tag!r} not found", fg=0xF00))
        note.labels.add(label)

    # Beam it up to Google's servers
    keep.sync()



from typing import Optional
from typing import *
import colr
import inquirer
from os import path

def dye(
    text: str,
    fg: Optional[int] = None,
    bg: Optional[int] = None,
    style: Optional[str] = None,
    no_closing: bool = False,
):
    return colr.color(
        text=text,
        fore=f"{fg:x}" if fg is not None else None,
        back=f"{bg:x}" if bg is not None else None,
        style=style,
        no_closing=no_closing,
    )


def ask(*questions) -> Union[List[Any], Any]:
    if len(questions) == 1:
        # No need to turn the hash into a tuple, just return the only value
        questions[0].name = "ans"
        return inquirer.prompt(questions)["ans"]

    for i, _ in enumerate(questions):
        questions[i].name = i
    # Ask dem questions
    answers = inquirer.prompt(questions)
    # Turn into a list of tuple
    answers = list(answers.items())
    # Sort by key
    answers = sorted(answers, key=lambda a: a[0])
    # Get only the answers
    answers = [a[1] for a in answers]
    return answers

def get_token_cache_filepath(service: str) -> str:
    return path.join(path.dirname(path.dirname(__file__)), f'.auth-cache--{service}')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ideaseed']

package_data = \
{'': ['*']}

install_requires = \
['colr>=0.9.1,<0.10.0',
 'docopt>=0.6.2,<0.7.0',
 'gkeepapi>=0.11.16,<0.12.0',
 'inquirer>=2.7.0,<3.0.0',
 'pygithub>=1.51,<2.0']

entry_points = \
{'console_scripts': ['ideaseed = ideaseed.cli:run']}

setup_kwargs = {
    'name': 'ideaseed',
    'version': '0.1.0',
    'description': 'Note down your ideas and get them to the right place, without switching away from your terminal',
    'long_description': '# ideaseed\n\n---\n\n_Disclaimer_ This project is at a "I\'m figuring out the interface" state, so nothing is working yet.\n\n---\n\nDo you have ideas suddenly and just wished you could catch them as fast as possible, as to not loose them, without having to interrupt what you were doing?\n\nAs I guy without a lot of more or less stupid ideas, I use Google Keep as a centralized place to put all of my thoughts that I deem worthy of consideration.\n\nI recently started to use GitHub Projects for _coding_ project ideas as a [single project called "incubator" on my GitHub profile directly](https://github.com/ewen-lbh?tab=projects), and as issues or notes when the idea is related to an already-existing project and repo.\n\nBut when I don\'t get to decide _when_ this idea comes, and I often need to interrupt what am I doing to open github, get to the right page, input my idea and get back. And I find it frustrating.\n\nEnough rambling. Here\'s what you came for.\n\nNote down your ideas and get them to the right place, without switching away from your terminal\n\n## Usage\n\n```bash\nideaseed (--help | --about | --version)\nideaseed [options] ARGUMENTS...\n```\n\n### Examples\n\n```sh-session\n# Save a card "test" in schoolsyst/webapp > project "UX" > column "To-Do"\n$ ideaseed schoolsyst/webapp UX "test"\n# Save a card "lorem" in your-username/ipsum > project "ipsum" > column "To-Do"\n$ ideaseed ipsum "lorem"\n# Save a card "a CLI to note down ideas named ideaseed" in your user profile > project "incubator" > column "willmake"\n$ ideaseed --user-keyword=project --user-project=incubator project "a CLI to note down ideas named ideaseed"\n```\n\n### Arguments\n\n| Argument | Meaning                                                                                                              | Default value  |\n| -------- | -------------------------------------------------------------------------------------------------------------------- | -------------- |\n| REPO     | Select a repository by name                                                                                          |\n|          | If not given, uses Google Keep instead of GitHub (or uses your user profile\'s projects if --project is used)         |\n|          | If --user-keyword\'s value is given, creates a card on your user\'s project (select which project with --user-project) |\n|          | If given in the form OWNER/REPO, uses the repository OWNER/REPO                                                      |\n|          | If given in the form REPO, uses the repository "your username/REPO"                                                  |\n| PROJECT  | Select a project by name to put your card to [default: REPO\'s value]                                                 | `REPO`\'s value |\n|          | If creating a card on your user\'s project, this becomes the COLUMN                                                   |\n| COLUMN   | Select a project\'s column by name [default: To-Do]                                                                   | To-Do          |\n|          | If creating a card on your user\'s project, this is ignored                                                           |\n\n### Options\n\n| Shorthand | Full-length         | Description                                                                                                  |\n| --------- | ------------------- | ------------------------------------------------------------------------------------------------------------ |\n| -c        | --color COLOR       | Chooses which color to use for Google Keep cards. See [Color names](#color-names) for a list of valid values |\n| -t        | --tag TAG           | Adds tags to the Google Keep card.                                                                           |\n| -i        | --issue TITLE       | Creates an issue with title TITLE.                                                                           |\n| -I        | --interactive       | Prompts you for the above options when they are not provided.                                                |\n| -L        | --logout            | Clears the authentification cache                                                                            |\n| -m        | --create-missing    | Create non-existant tags, projects or columns specified (needs confirmation if -I is used)                   |\n|           | --about             | Details about ideaseed like currently-installed version                                                      |\n|           | --version           | Like --about, without dumb and useless stuff                                                                 |\n|           | --user-project NAME | Name of the project to use as your user project                                                              |\n|           | --user-keyword NAME | When REPO is NAME, creates a GitHub card on your user profile instead of putting it on REPO                  |\n|           | --no-auth-cache     | Don\'t save credentials in a temporary file                                                                   |\n\n#### Color names\n\n- blue\n- brown\n- darkblue\n- gray\n- green\n- orange\n- pink\n- purple\n- red\n- teal\n- white\n- yellow\n\nYou don\'t have to specify the whole color name, just enough to be non-ambiguous:\n\n- bl\n- br\n- d\n- gra\n- gre\n- o\n- pi\n- pu\n- r\n- t\n- w\n- y\n\n#### Relax. You don\'t need to remember those options\n\nYou can also use `ideaseed -I` to prompt you for some information:\n\n- Where do you want to upload this idea? (github, google keep)\n- If you decide to use github,\n  - On your profile?\n  - If not:\n    - Which repo? (using REPO or OWNER/REPO) (autocompletes with repositories you contribute to)\n    - Which column? (choices are the column names, and you can type the column\'s index to be quicker)\n- If you decide to use google keep,\n  - Which color? (defaults to white)\n  - Some tags?\n',
    'author': 'Ewen Le Bihan',
    'author_email': 'ewen.lebihan7@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ewen-lbh/ideaseed',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
